#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include <stdint.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <set>
#include <vector>
#include <algorithm>
#include <random>



static const unsigned NO_COST = 0;
static const unsigned COMBINED = 1;
static const unsigned ECB_UNION = 2;
static const unsigned UCB_UNION = 3;
static const unsigned UCB_ONLY = 4;
static const unsigned ECB_ONLY = 5;
static const unsigned CAROUSEL = 6;
static const unsigned EXCEPTION1 = 7;
static const unsigned EXCEPTION2 = 8;
static const unsigned SPMB = 9;
static const unsigned COUNTER_SIZE = 10;


static const unsigned MAX_THREADS = 16;
static const unsigned MAX_TASKS = 30;
static const unsigned DEFAULT_NUM_TASKS = 15;
static const unsigned SEED = 2;

#include "spm_data.h"

static const double EPSILON = 1e-5;

static uint8_t byte_bit_count[0x100];

static void init_byte_bit_count()
{
    unsigned i, j;

    for (i = 0; i < 0x100; i++) {
        byte_bit_count[i] = 0;
        for (j = 1; j < 0x100; j*=2) {
            if (i & j) {
                byte_bit_count[i]++;
            }
        }
    }
}


class Cache_Block_Set
{
public:
    Cache_Block_Set(unsigned _num_cache_sets) {
        init(_num_cache_sets);
    }

    void init(unsigned _num_cache_sets) {
        num_cache_sets = _num_cache_sets;
        num_words = (num_cache_sets + 31) / 32;
        total = 0;
        dirty = false;
        contents = new uint32_t[num_words];
        clear();
        assert(byte_bit_count[0xff] == 8);
        assert(byte_bit_count[0x77] == 6);
    }

    void operator=(const Cache_Block_Set & orig) {
        delete[] contents;
        init(orig.num_cache_sets);
        memcpy(contents, orig.contents, num_words * 4);
        dirty = true;
    }

    Cache_Block_Set(const Cache_Block_Set & orig) {
        init(orig.num_cache_sets);
        memcpy(contents, orig.contents, num_words * 4);
        dirty = true;
    }

    ~Cache_Block_Set() {
        delete[] contents;
    }

    void clear(void) {
        total = 0;
        dirty = false;
        for (unsigned i = 0; i < num_words; i++) {
            contents[i] = 0;
        }
    };

    void insert(unsigned bit) {
        assert(bit < num_cache_sets);
        contents[bit / 32] |= 1 << (bit % 32);
        dirty = true;
    };

    bool test(unsigned bit) {
        assert(bit < num_cache_sets);
        return (contents[bit / 32] & (1 << (bit % 32))) != 0;
    };

    unsigned size() {
        unsigned i, j;
        if (!dirty) {
            return total;
        }
        total = 0;
        for (i = 0; i < num_words; i++) {
            uint32_t word = contents[i];
            for (j = 0; j < 4; j++) {
                total += byte_bit_count[(word >> (j * 8)) & 0xff];
            }
        }
        dirty = false;
        return total;
    };

    unsigned first() {
        unsigned bit;
        for (bit = 0; bit < num_cache_sets; bit++) {
            if (contents[bit / 32] & (1 << (bit % 32))) {
                return bit;
            }
        }
        return num_cache_sets;
    }

    void union_ip(Cache_Block_Set & other) {
        assert(other.num_cache_sets == num_cache_sets);
        for (unsigned i = 0; i < num_words; i++) {
            contents[i] |= other.contents[i];
        }
        dirty = true;
    };

    void intersect_ip(Cache_Block_Set & other) {
        assert(other.num_cache_sets == num_cache_sets);
        for (unsigned i = 0; i < num_words; i++) {
            contents[i] &= other.contents[i];
        }
        dirty = true;
    };

private:
    unsigned num_cache_sets, total, num_words;
    uint32_t * contents;
    bool dirty;

    
};


class Task
{
public:
    Task(unsigned num_cache_sets, const char * _name) : 
            Ccache(0), T(0), R(0), Cspm(0), Sspm(0), Sinitial_load(0),
            Csave(0), Crestore(0), Bspm(0), Bcache(0),
            UCB(num_cache_sets), ECB(num_cache_sets),
            name(_name), blocking_rule(0) {};

    void print(FILE * fd);

    double Ccache, T, R, Cspm;
    unsigned Sspm, Sinitial_load;
    double Csave, Crestore, Bspm, Bcache;
    Cache_Block_Set UCB;
    Cache_Block_Set ECB;
    std::string name;
    unsigned blocking_rule;
};

void Task::print(FILE * fd)
{
    fprintf(fd, "\t%s\n", name.c_str());
    fprintf(fd, "\tCcache = %1.0fns\n", 1e9 * Ccache);
    fprintf(fd, "\tCspm = %1.0fns\n", 1e9 * Cspm);
    fprintf(fd, "\tT = %1.0fns\n", 1e9 * T);
    fprintf(fd, "\tCsave = %1.0fns\n", 1e9 * Csave);
    fprintf(fd, "\tCrestore = %1.0fns\n", 1e9 * Crestore);
    fprintf(fd, "\tBcache = %1.0fns\n", 1e9 * Bcache);
    fprintf(fd, "\tBspm = %1.0fns\n", 1e9 * Bspm);
    fprintf(fd, "\tblocking_rule = %u\n", blocking_rule);
    fprintf(fd, "\tSspm = %u\n", Sspm);
    fprintf(fd, "\tSinitial_load = %u\n", Sinitial_load);
    fprintf(fd, "\t|UCB| = %u\n", UCB.size());
    fprintf(fd, "\t|ECB| = %u\n", ECB.size());
    fprintf(fd, "\tR = %1.0fns\n", 1e9 * R);
    if (R <= T) {
        fprintf(fd, "\tschedulable\n");
    }
}

class Task_Set
{
public:
    Task_Set(unsigned seed) {
        num_cache_sets = 2048 / 16;
        total_cache_utilisation = 10.0;
        reuse_factor = 0.3;
        num_tasks = DEFAULT_NUM_TASKS;
        S_factor = -1.0; 
        equal_c = blocking_alternate = brt_same = false;
        min_c = min_blocks = alt_task_selection = false;
        synth = same_blocks = false;
        true_spm = true;
        override_max_size = 0;
        cache_usage = 0;
        spm_usage = 0;

        brt_cache = BRT_CACHE;
        brt_spm = BRT_SPM;
        b_cache = B_CACHE;
        b_spm = B_SPM;

        memset(buffer, 0, sizeof(buffer));
        srand48_r(seed, &random_state);
        reset();
    }

    void reset();
    bool response_times(unsigned mm);
    double calc_UCB_only(unsigned i, unsigned j);
    double calc_UCB_union(unsigned i, unsigned j);
    double calc_ECB_union(unsigned i, unsigned j);
    double drandom(void);
    unsigned irandom(unsigned n);
    std::vector<double> uunifast(unsigned num_tasks, double sumU);
    void task_set_generator(double total_utilisation);
    void test_once(bool complete);
    void test_new(const char * fname, double u);
    void create_gamma_table(double * gamma_table, unsigned mm);
    void idealise();
    void print(FILE * fd);

    double S_factor;
    unsigned num_cache_sets;
    double total_cache_utilisation, reuse_factor;
    bool equal_c, blocking_alternate, brt_same, same_blocks;
    bool alt_task_selection, true_spm, min_blocks, min_c, synth;
    unsigned override_max_size;
    unsigned num_tasks;
    std::vector<Task> tasks;
    struct drand48_data random_state;
    unsigned counter[COUNTER_SIZE];
    double W_value[COUNTER_SIZE];
    char buffer[BUFSIZ];
    unsigned * cache_usage;
    unsigned * spm_usage;

    double brt_cache, brt_spm, b_cache, b_spm;
};


void Task_Set::reset()
{
    memset(counter, 0, sizeof(counter));
}

void Task_Set::idealise()
{
    brt_spm = brt_cache;

    min_blocks = true;
    same_blocks = true;
    equal_c = true;
    min_c = true;
    brt_same = true;
}

void Task_Set::print(FILE * fd)
{
    unsigned i;
    for (i = 0; i < (unsigned) tasks.size(); i++) {
        fprintf(fd, "Task %u {\n", i);
        tasks[i].print(fd);
        fprintf(fd, "}\n");
    }
}

void Task_Set::create_gamma_table(double * gamma_table, unsigned mm)
{
    const unsigned num_tasks = tasks.size();
    unsigned i, j;

    for (i = 0; i < num_tasks; i++) {
        for (j = 0; j < i; j++) {
            double gamma = 0.0;

            switch(mm) {
            case CAROUSEL:  
                gamma = tasks[j].Csave + tasks[j].Crestore;
                break;
            case NO_COST:   
                break;
            case ECB_ONLY:  
                gamma = brt_cache * tasks[j].ECB.size();
                break;
            case UCB_ONLY:  
                gamma = calc_UCB_only(i, j);
                break;
            case UCB_UNION: 
                gamma = calc_UCB_union(i, j);
                break;
            case ECB_UNION: 
                gamma = calc_ECB_union(i, j);
                break;
            default:        
                assert(false);
                break;
            }
            gamma += CS1 + CS2;
            gamma_table[(i * num_tasks) + j] = gamma;
        }
        for (; j < num_tasks; j++) {
            gamma_table[(i * num_tasks) + j] = 0.0;
        }
    }
}

bool Task_Set::response_times(unsigned mm)
{
    unsigned i, j;
    bool success = true;
    const unsigned num_tasks = tasks.size();
    double * gamma_table = (double *) 
                alloca((num_tasks * num_tasks) * sizeof(double));

    create_gamma_table(gamma_table, mm);

    for (i = 0; i < num_tasks; i++) {
        double Ci, Bi, base;
        double Ti = tasks[i].T;

        if (mm != CAROUSEL) {
            // Cache
            Ci = tasks[i].Ccache;
            Bi = tasks[i].Bcache;
            base = Ci + CS1 + Bi;
        } else {
            // MSRS
            Ci = tasks[i].Cspm;
            Bi = tasks[i].Bspm;
            base = Ci + CS1 + Bi + tasks[i].Csave;
        }


        double Ri1 = base;
        double Ri0 = 0.0;
        bool schedulable = true;

        while (fabs(Ri1 - Ri0) > EPSILON) {
            Ri0 = Ri1;
            Ri1 = base;

            for (j = 0; j < i; j++) {
                double gamma = gamma_table[(i * num_tasks) + j];
                double Cj = tasks[j].Ccache;
                double t1;

                if (mm == CAROUSEL) {
                    Cj = tasks[j].Cspm;
                }
                
                t1 = ceil(Ri0 / tasks[j].T);
                assert((Cj + gamma) >= 0);
                assert(t1 >= 0);
                Ri1 += t1 * (Cj + gamma);
            }
            assert((Ri1 - Ri0) >= (-EPSILON));
            if (Ri1 > Ti) {
                // Deadline miss - not schedulable
                schedulable = false;
                break;
            }
        }
        
        if (schedulable) {
            tasks[i].R = Ri1;
        } else {
            tasks[i].R = Ti * 2.0;
            success = false;
        }
    }
    return success;
}

double Task_Set::calc_UCB_only(unsigned i, unsigned j)
{
    double x = 0.0;
    unsigned k;

    for (k = j + 1; k < (i + 1); k++) {
        double y = tasks[k].UCB.size();
        if (y > x) {
            x = y;
        }
    }
    return brt_cache * x;
}

double Task_Set::calc_UCB_union(unsigned i, unsigned j)
{
    Cache_Block_Set all_ucbs(num_cache_sets);
    unsigned k;

    for (k = j + 1; k < (i + 1); k++) {
        all_ucbs.union_ip(tasks[k].UCB);
    }

    all_ucbs.intersect_ip(tasks[j].ECB);
    return brt_cache * all_ucbs.size();
}

double Task_Set::calc_ECB_union(unsigned i, unsigned j)
{
    unsigned k, h;
    double x;
    Cache_Block_Set all_ucbs(num_cache_sets);

    x = 0;
    for (k = j + 1; k < (i + 1); k++) {
        all_ucbs.clear();

        for (h = 0; h < (j + 1); h++) {
            all_ucbs.union_ip(tasks[h].ECB);
        }
        all_ucbs.intersect_ip(tasks[k].UCB);
        double y = brt_cache * all_ucbs.size();
        if (y > x) {
            x = y;
        }
    }
    return x;
}

static bool deadline_monotonic(const Task & a, const Task & b) 
{
    return a.T < b.T;
}

std::vector<double> Task_Set::uunifast(unsigned num_tasks, 
                double total_utilisation)
{
    std::vector<double> out;
    double sumU = total_utilisation;
    unsigned i;

    for (i = 1; i < num_tasks; i++) {
        double x = drandom();
        double y = 1.0 / (double) (num_tasks - i);
        double nextSumU = sumU * pow(x, y);
        out.push_back(sumU - nextSumU);
        sumU = nextSumU;
    }
    out.push_back(sumU);
    assert(out.size() == num_tasks);

    double total_out = 0.0;
    for (i = 0; i < num_tasks; i++) {
        assert(out[i] >= 0);
        total_out += out[i];
    }
    assert(fabs(total_out - total_utilisation) < 1e-6);
    return out;
}

double Task_Set::drandom(void)
{
    double out;
    drand48_r(&random_state, &out);
    return out;
}

unsigned Task_Set::irandom(unsigned n)
{
    long int j;
    lrand48_r(&random_state, &j);
    j %= n;
    return j;
}

static void check_mrtc_table()
{
    unsigned i;
    for (i = 0; i < NUM_MRTC_TASKS; i++) {
        double x = mrtc_table[i].perfect_wcet + 
                (mrtc_table[i].num_ecbs * BRT_CACHE);
            
        assert(fabs(mrtc_table[i].cache_wcet - x) < 1e-8);

        x = mrtc_table[i].perfect_wcet +
                (mrtc_table[i].num_blocks * BRT_SPM) +
                (mrtc_table[i].num_regions * TRANSITION_WCET_OVERHEAD);

        assert(fabs(mrtc_table[i].spm_wcet - x) < 1e-8);
    }
}

void Task_Set::task_set_generator(double total_utilisation)
{
    unsigned chosen_tasks[MAX_TASKS];
    unsigned i, j;

    assert(NUM_MRTC_TASKS <= MAX_TASKS);
    assert(num_tasks <= MAX_TASKS);
    assert(num_tasks > 0);

    memset(chosen_tasks, 0x80, sizeof(chosen_tasks));

    unsigned max_size = num_cache_sets;

    if (override_max_size) {
        max_size = override_max_size;
    }

    // Uniform selection from the MRTC sample
    unsigned ok = 0;
    for (i = 0; i < NUM_MRTC_TASKS; i++) {
        if ((mrtc_table[i].num_ecbs <= max_size)
        && (mrtc_table[i].num_blocks <= max_size)) {
            ok ++;
        }
    }
    assert(ok >= 3);

    for (i = 0; i < num_tasks; i++) {
        j = irandom(NUM_MRTC_TASKS);

        // Re-roll if you pick one that's too big
        while ((mrtc_table[j].num_ecbs > max_size)
        || (mrtc_table[j].num_blocks > max_size)) {
            j = irandom(NUM_MRTC_TASKS);
        }
        chosen_tasks[i] = j;
    }

    // choose num_tasks tasks
    tasks.clear();
    for (i = 0; i < num_tasks; i++) {
        assert(chosen_tasks[i] < NUM_MRTC_TASKS);
        tasks.push_back(Task(num_cache_sets,
                mrtc_table[chosen_tasks[i]].name));
    }
    std::vector<double> utilisations;
        
    utilisations = uunifast(num_tasks, total_utilisation);

    for (i = 0; i < num_tasks; i++) {
        unsigned j = chosen_tasks[i];
        double Cbase = mrtc_table[j].perfect_wcet;
        unsigned num_blocks = mrtc_table[j].num_blocks;
        unsigned num_regions = mrtc_table[j].num_regions;
        unsigned Sspm = mrtc_table[j].S_spm;

        if (S_factor >= 0.0) {
            num_regions = 1;
            num_blocks = mrtc_table[j].num_ecbs;
            Sspm = mrtc_table[j].num_ucbs + (unsigned)
                    floor(S_factor * (double) (mrtc_table[j].num_ecbs - 
                                mrtc_table[j].num_ucbs));
        }

        // C^\cache_i (Load and Execute)
        tasks[i].Ccache = (mrtc_table[j].num_ecbs * brt_cache) + Cbase;
        tasks[i].Bcache = b_cache;

        // C^\spm_i (Load and Execute)
        tasks[i].Cspm = (num_blocks * brt_spm) + Cbase +
                (num_regions * TRANSITION_WCET_OVERHEAD);
       
        tasks[i].Csave = (SAVE_PER_BLOCK * Sspm) + SAVE_OVERHEAD -
                TRANSITION_WCET_OVERHEAD;
        tasks[i].Crestore = (brt_spm * Sspm) + RESTORE_OVERHEAD;

        tasks[i].Sspm = Sspm;
        assert(tasks[i].Sspm <= max_size);

        tasks[i].Sinitial_load = mrtc_table[j].S_initial_load;
        assert(tasks[i].Sinitial_load <= Sspm);
        assert(0 < tasks[i].Sinitial_load);

        tasks[i].T = tasks[i].Ccache / utilisations[i];
    }

    // Populate ECB and UCB
    unsigned startSet = irandom(num_cache_sets);
    for (i = 0; i < num_tasks; i++) {
        unsigned j = chosen_tasks[i];
        unsigned ecbsNr = mrtc_table[j].num_ecbs;
        unsigned ucbsNr = mrtc_table[j].num_ucbs;

        assert(ecbsNr <= max_size);
        assert(ecbsNr != 0);
        assert(ucbsNr <= ecbsNr);

        for (j = 0; j < ecbsNr; j++) {
            tasks[i].ECB.insert((j + startSet) % num_cache_sets);

            if (j < ucbsNr) {
                tasks[i].UCB.insert((j + startSet) % num_cache_sets);
            }
        }
        startSet = (startSet + ecbsNr) % num_cache_sets;
    }

    // tasks sorted into deadline monotonic order
    std::sort(tasks.begin(), tasks.end(), deadline_monotonic);
    assert(tasks[num_tasks-1].T >= tasks[0].T);

    // Calculate B^\spm_i 
    for (i = 0; i < num_tasks; i++) {
        double blocker;
        tasks[i].Bspm = b_cache;
        tasks[i].blocking_rule = 0;

        // Blocking from lower-priority tasks
        for (j = i + 1; j < num_tasks; j++) {
            // task startup blocking
            // Caused by CS1, Save, or initial Load
            blocker = CS1 + tasks[j].Csave + (brt_spm * tasks[j].Sinitial_load)
                                + TRANSITION_WCET_OVERHEAD;
            if (tasks[i].Bspm < blocker) {
                tasks[i].Bspm = blocker;
                tasks[i].blocking_rule = 1;
            }

            // task within blocking
            // Caused by some Load (maximum size of Load is Sspm)
            blocker = (tasks[j].Sspm * brt_spm) + TRANSITION_WCET_OVERHEAD;
            if (tasks[i].Bspm < blocker) {
                tasks[i].Bspm = blocker;
                tasks[i].blocking_rule = 2;
            }

            // task end blocking
            // Caused by CS2 or Restore
            blocker = CS2 + tasks[j].Crestore;
            if (tasks[i].Bspm < blocker) {
                tasks[i].Bspm = blocker;
                tasks[i].blocking_rule = 3;
            }
        }
        if (blocking_alternate) {
            // Assume Save, Restore etc are interruptible
            tasks[i].Bspm = b_cache;
            tasks[i].blocking_rule = 4;
        }

        // Blocking from task i's restore time
        blocker = CS2 + tasks[i].Crestore;
        if (tasks[i].Bspm < blocker) {
            tasks[i].Bspm = blocker;
            tasks[i].blocking_rule = 5;
        }
    }
}


void Task_Set::test_once(bool complete)
{
    unsigned i;
    std::vector<double> ecb_r;

    if (complete) {
        counter[NO_COST] +=     response_times(NO_COST) ? 1 : 0;
        counter[UCB_ONLY] +=    response_times(UCB_ONLY) ? 1 : 0;
        counter[ECB_ONLY] +=    response_times(ECB_ONLY) ? 1 : 0;
    }

    counter[ECB_UNION] +=   response_times(ECB_UNION) ? 1 : 0;
    for (i = 0; i < tasks.size(); i++) {
        ecb_r.push_back(tasks[i].R);
    }

    counter[UCB_UNION] +=   response_times(UCB_UNION) ? 1 : 0;

    bool success = true;
    for (i = 0; i < tasks.size(); i++) {
        double a = ecb_r[i];
        double b = tasks[i].R;
        if (b < a) {
            a = b;
        }
        ecb_r[i] = a;
        tasks[i].R = a;
        if (a > tasks[i].T) {
            success = false;
        }
    }
    counter[COMBINED] +=    success ? 1 : 0;

    if (complete) {
        bool csel = response_times(CAROUSEL);
        counter[CAROUSEL] +=    csel ? 1 : 0;

        if ((!success) && csel) {
            counter[EXCEPTION1] ++;
            //printf("exception\n");
        } else if (success && !csel) {
            counter[EXCEPTION2] ++;
        }

        bool spm_better = false;
        bool spm_always_better = true;
        for (i = 0; i < tasks.size(); i++) {
            if (ecb_r[i] >= tasks[i].R) {
                spm_better = true;
            } else {
                spm_always_better = false;
            }
        }
        if (spm_always_better) {
            counter[SPMB] ++;
        }
    }
}


static pthread_mutex_t io_critical_section;
static unsigned done;
static char output_dir[BUFSIZ];
static unsigned num_runs;

void Task_Set::test_new(const char * fname, double u)
{
    const unsigned CRPD = 0;
    const unsigned SRPD = 1;
    const unsigned SPECIAL_SRPD = 2;
    const unsigned SPECIAL_CRPD = 3;
    const unsigned NUM_COUNTERS = 4;
    unsigned counter2[NUM_COUNTERS];
    unsigned i;
    char fname2[BUFSIZ];
    int rc;

    rc = snprintf(fname2, BUFSIZ - 1, "%s/%s", output_dir, fname);
    if ((rc <= 0) || (rc >= (BUFSIZ - 1))) {
        perror("truncated snprintf");
        return;
    }

#if 0
    struct stat s;
    if (stat(fname2, &s) == 0) {
        if (!(S_IWUSR & s.st_mode)) {
            // read-only file, skip tests
            return;
        }
    }
#endif

    memset(counter2, 0, sizeof(counter2));

    for (i = 0; i < num_runs; i++) {
        reset();
        task_set_generator(u);
        test_once(false);

        bool crpd_yes = (counter[COMBINED] != 0);
        bool srpd_yes = response_times(CAROUSEL);

        if (srpd_yes && !crpd_yes) {
            counter2[SPECIAL_SRPD]++;
        }
        if (crpd_yes && !srpd_yes) {
            counter2[SPECIAL_CRPD]++;
        }
        if (srpd_yes) {
            counter2[SRPD]++;
        }
        if (crpd_yes) {
            counter2[CRPD]++;
        }
    }

    pthread_mutex_lock(&io_critical_section);
    {
        FILE * fd;
        fd = fopen(fname2, "at");
        if (!fd) {
            perror("fd == NULL");
            fprintf(stderr, "filename: %s\n", fname2);
        } else {
            fprintf(fd, "%-30s %1.2f ", fname, u);
            printf("%-30s %1.2f ", fname, u);
            for (i = 0; i < NUM_COUNTERS; i++) {
                fprintf(fd, " %7u", counter2[i]);
                printf(" %7u", counter2[i]);
            }
            printf("\n");
            fprintf(fd, "\n");
            fclose(fd);
            fflush(stdout);
        }
    }
    pthread_mutex_unlock(&io_critical_section);
}

struct Thread_Instruction {
    double start, delta;
    pthread_t handle;
    unsigned thread_id, num_threads;
};

/////////////////////////////////////////////////////////////////////
// Results for section 4 - the advantage of MSRS
void * section_4(void * data)
{
    Thread_Instruction *    ti = (Thread_Instruction *) data;
    double                  u;
    Task_Set *              ts;

    for (u = ti->start; u <= (1.0 + EPSILON); u += ti->delta) {
        ts = new Task_Set(SEED);
        ts->test_new("real.csv", u);    // Figure 3
        delete ts;

        ts = new Task_Set(SEED);
        ts->S_factor = 0.0;
        ts->test_new("good.csv", u); // Figure 2 (good SRPD)
        delete ts;

        ts = new Task_Set(SEED);
        ts->S_factor = 1.0;
        ts->test_new("poor.csv", u); // Figure 2 (bad SRPD)
        delete ts;

        ts = new Task_Set(SEED);
        ts->S_factor = 0.0;
        ts->blocking_alternate = true;
        ts->test_new("loblock.csv", u);
        delete ts;
    }
    return 0;
}


/////////////////////////////////////////////////////////////////////
// Results for section 5.1 - number of tasks
void * section_5_1(void * data)
{
    Thread_Instruction *    ti = (Thread_Instruction *) data;
    double                  u;
    Task_Set *              ts;
    char                    buf[BUFSIZ];
    unsigned                i;

    for (u = ti->start; u <= (1.0 + EPSILON); u += ti->delta) {
        for (i = 1; i <= MAX_TASKS; i++) {
            ts = new Task_Set(SEED);
            ts->num_tasks = i;
            ts->S_factor = 0.0;
            sprintf(buf, "ntasks%02u.ideal.csv", i);
            ts->test_new(buf, u);
            delete ts;
        }
    }
    return 0;
}

/////////////////////////////////////////////////////////////////////
// Results for section 5.2 - BRT ratio
void * section_5_2(void * data)
{
    Thread_Instruction *    ti = (Thread_Instruction *) data;
    double                  u;
    Task_Set *              ts;
    char                    buf[BUFSIZ];
    unsigned                i;

    for (u = ti->start; u <= (1.0 + EPSILON); u += ti->delta) {
        for (i = 70; i <= 130; i++) {
            double f = ((double) i) * 0.01;
            ts = new Task_Set(SEED);
            ts->brt_cache = BRT_CACHE;
            ts->brt_spm = (double) BRT_CACHE * f;
            ts->S_factor = 0.0;
            sprintf(buf, "brt%02u.ideal.csv", i);
            ts->test_new(buf, u);
            delete ts;
        }
    }
    return 0;
}

/////////////////////////////////////////////////////////////////////
// Results for section 5.3 - S^spm to \ECB ratio
void * section_5_3(void * data)
{
    Thread_Instruction *    ti = (Thread_Instruction *) data;
    double                  u;
    Task_Set *              ts;
    char                    buf[BUFSIZ];
    unsigned                i;

    for (u = ti->start; u <= (1.0 + EPSILON); u += ti->delta) {
        for (i = 0; i <= 20; i++) {
            double f = ((double) i) * 0.05;
            ts = new Task_Set(SEED);
            ts->S_factor = f;
            sprintf(buf, "secb%02u.ideal.csv", i);
            ts->test_new(buf, u);
            delete ts;
        }
    }
    return 0;
}

/////////////////////////////////////////////////////////////////////
// Results for section 5.4 - local memory size
void * section_5_4(void * data)
{
    Thread_Instruction *    ti = (Thread_Instruction *) data;
    double                  u;
    Task_Set *              ts;
    char                    buf[BUFSIZ];
    unsigned                i;

    for (u = ti->start; u <= (1.0 + EPSILON); u += ti->delta) {
        for (i = 500; i <= 1200; i += 10) {
            ts = new Task_Set(SEED);
            ts->num_cache_sets = (unsigned) floor(pow(2.0, i / 100.0));
            ts->S_factor = 0.0;
            //ts->override_max_size = 1 << 12;
            sprintf(buf, "localmem%04u.ideal.csv", i);
            ts->test_new(buf, u);
            delete ts;
        }
    }
    return 0;
}

/////////////////////////////////////////////////////////////////////
// Results for section 5.5 - blocking introduced
void * section_5_5(void * data)
{
    Thread_Instruction *    ti = (Thread_Instruction *) data;
    double                  u;
    Task_Set *              ts;
    char                    buf[BUFSIZ];
    unsigned                i;

    for (u = ti->start; u <= (1.0 + EPSILON); u += ti->delta) {
        for (i = 0; i <= 40; i ++) {
            ts = new Task_Set(SEED);
            ts->num_tasks = 10;
            ts->idealise();
            ts->b_spm = (B_SPM * (double) i) / 40.0;
            sprintf(buf, "blocking%02u.ideal.csv", i);
            ts->test_new(buf, u);
            delete ts;
        }
    }
    return 0;
}

/////////////////////////////////////////////////////////////////////
// Results for C^spm versus C^cache comparison
void * compare_cc(void * data)
{
    Thread_Instruction *    ti = (Thread_Instruction *) data;
    double                  u;
    Task_Set *              ts;
    char                    buf[BUFSIZ];
    unsigned                i;

    for (u = ti->start; u <= (1.0 + EPSILON); u += ti->delta) {
        for (i = 5; i <= 20; i++) {
            double f = ((double) i) * 0.1;
            ts = new Task_Set(SEED);
            ts->brt_cache = BRT_CACHE;
            ts->brt_spm = (double) BRT_CACHE * f;
            ts->num_tasks = 10;
            ts->idealise();
            ts->equal_c = false;
            ts->min_c = false;
            ts->brt_same = true;
            sprintf(buf, "cc%02u.ideal.csv", i);
            ts->test_new(buf, u);
            delete ts;
        }
    }
    return 0;
}

void check(void)
{
    Task_Set * ts = new Task_Set(1);
    ts->num_tasks = 5;
    ts->S_factor = 0.0;
    ts->reset();
    ts->task_set_generator(0.4);
    ts->test_once(false);
    ts->response_times(CAROUSEL);
    ts->print(stdout);
    delete ts;
}

int main(int argc, char ** argv)
{
    Thread_Instruction ti[MAX_THREADS];
    unsigned i;
    bool use_output_dir = true;
    void * (* thread_proc) (void *) = 0;

    init_byte_bit_count();
    check_mrtc_table();
    done = 0;

#if 0
    for (unsigned s = 0; s < 10; s++) {
        Task_Set * ts = new Task_Set(s + 100);
        ts->num_tasks = 10;
        ts->reset();
        ts->task_set_generator(0.6);

        for (i = 0; i < ts->num_tasks; i++) {
            printf("%10.8f %10.8f  %10.8f %10.8f  %10.8f %s\n",
                ts->tasks[i].Ccache,
                ts->tasks[i].Bcache,
                ts->tasks[i].Cspm,
                ts->tasks[i].Bspm,
                ts->tasks[i].T,
                ts->tasks[i].name.c_str());
        }
        //ts->test_once(false);
        //ts->test_new("srpdfig2.csv", u);    // Figure 3
        delete ts;
        printf("\n\n");
    }
    return 0;
#endif

    if (argc != 5) {
        fprintf(stderr, "usage: %s <output dir> "
                    "<num runs> <num threads> <mode>\n"
                 "mode [1-5] - paper section 5 (subsections)\n"
                 "mode q - paper section 4\n"
                 "mode c - C^spm/C^cache comparison\n"
                 "",
                argv[0]);
        return 1;
    }

    strncpy(output_dir, argv[1], BUFSIZ - 1);
    output_dir[BUFSIZ - 1] = '\0';
    num_runs = atoi(argv[2]);
    if (num_runs <= 0) {
        fprintf(stderr, "invalid number of runs\n");
        return 1;
    }
    unsigned num_threads = atoi(argv[3]);
    if (sysconf(_SC_NPROCESSORS_ONLN) < num_threads) {
        num_threads = sysconf(_SC_NPROCESSORS_ONLN);
    }
    if ((num_threads == 0) || (num_threads > MAX_THREADS)) {
        fprintf(stderr, "invalid number of threads\n");
        return 1;
    }

    switch(tolower(argv[4][0])) {
    case 'q':   thread_proc = &section_4;
                break;
    case '1':   thread_proc = &section_5_1;
                break;
    case '2':   thread_proc = &section_5_2;
                break;
    case '3':   thread_proc = &section_5_3;
                break;
    case '4':   thread_proc = &section_5_4;
                break;
    case '5':   thread_proc = &section_5_5;
                break;
    case 'x':   check();
                return 0;
    default:    fprintf(stderr, "Unknown mode: '%c'\n", argv[4][0]);
                return 1;
    }
    if (use_output_dir && (0 != mkdir(output_dir, 0755))) {
        perror("mkdir");
        fprintf(stderr, "unable to create output dir\n");
        return 1;
    }


    const double DELTA = 0.01;

    for (i = 0; i < num_threads; i++) {
        ti[i].start = DELTA * (double) i;
        ti[i].delta = DELTA * (double) num_threads;
        ti[i].thread_id = i;
        ti[i].num_threads = num_threads;
    }

    pthread_mutex_init(&io_critical_section, 0);


    for (i = 0; i < num_threads; i++) {
        if (0 != pthread_create(&ti[i].handle, 0,
                    thread_proc, &ti[i])) {
            perror("pthread_create");
            return 1;
        }
    }
    for (i = 0; i < num_threads; i++) {
        void * res;
        pthread_join(ti[i].handle, &res);
    }
    pthread_mutex_destroy(&io_critical_section);

    return 0;
}


