/* RTOS C-level code including the scheduler function */
/* Carousel and cache variants */

#include "rtos.h"

#define MAX_NAME_LEN        8

typedef struct TCB_struct {
        void *      function;   /* offset 0 for asm */
        unsigned    running;    /* offset 4 for asm */
        unsigned    block_info_word; /* offset 8 for asm */
        unsigned    dynamic_release, dynamic_deadline, period;
        unsigned    offset;
        char        name[MAX_NAME_LEN+1];
} TCB;

static TCB thread_table[MAX_THREADS] AKD;
static unsigned num_threads AKD = 0;

/*asm(".section .kernelstuff.t,\"ax\",@progbits\n");*/

void register_task(const char * name, 
                void * function, unsigned block_info_word,
                unsigned offset, unsigned period) 
{
    TCB * p = &thread_table[num_threads + 1];

    if ((num_threads + 1) >= MAX_THREADS) {
        print("too many threads\n");
        exit(EXIT_NO_TASK);
    }
    if (!function) {
        print("null function registered!\n");
        exit(EXIT_NO_TASK);
    }
    num_threads++;
    p->function = function;
    p->block_info_word = block_info_word;
    p->offset = offset;
    p->period = period;
    memcpy(p->name, name, MAX_NAME_LEN);
    p->name[MAX_NAME_LEN] = '\0';
    notify_task_name(name, function);
}

void * scheduler(void) 
{
    /* called by ISR */
    unsigned i;
    unsigned now = get_time();
    unsigned preempt = 0xfffffffe;
    TCB * best_task = &thread_table[0]; /* idle task, always runnable */

    if ((RUN_FOR) && (now > RUN_FOR)) {
        print("\nreached RUN_FOR limit\n");
        exit(EXIT_RUN_FOR);
    }

    /* Check deadlines */
    for (i = num_threads; i > 0; i--) {
        TCB * p = &thread_table[i];

        if (p->dynamic_deadline < now) {
            print("*** deadline overrun ***\n");
            for (i = num_threads; i > 0; i--) {
                p = &thread_table[i];
                if (p->dynamic_deadline < now) {
                    print(p->name);
                }
            }
            print("halted\n");
            exit(EXIT_DEADLINE_OVERRUN);
        }
    }


    /* 1. What is the best_task to run now? */
    /* 2. At what time should we preempt? */

    for (i = num_threads; i > 0; i--) {
        TCB * p = &thread_table[i];
        
        if (p->running) {
            /* Do not reschedule; no new task */
            best_task = p;
            break;
        }

        if (p->dynamic_release <= now) {
            /* Runnable task */
            best_task = p;
            break;
        } else if (p->dynamic_release < preempt) {
            /* Higher-priority task, not yet runnable */
            preempt = p->dynamic_release;
        }
    }

    if (best_task->running) {
        /* No task switch this time; schedule preemption */
        set_trigger(preempt);
        return 0;
    } else {
        /* A new task starts; schedule preemption if any */
        set_trigger(preempt);
        best_task->dynamic_release += best_task->period;
        best_task->dynamic_deadline += best_task->period;
        best_task->running = 1;
        return (void *) best_task;
    }
}

extern void kernel_idle(void);

void crash(void) 
{
    print("crash called; idle was rescheduled?");
    exit(EXIT_CRASH);
}

void kernel_start(void) 
{
    unsigned now = get_time();
    unsigned i;

    /* initiate idle task */
    thread_table[0].function = crash;
    thread_table[0].running = 1;

    /* setup release times and deadlines */
    for (i = num_threads; i > 0; i--) {
        TCB * p = &thread_table[i];
        p->dynamic_release = p->offset + now;
        p->dynamic_deadline = p->period + p->offset + now;
        p->running = 0;
    }

    /* start timer, enable interrupts and go idle */
    kernel_idle();
}

void printhex(unsigned x)
{
    unsigned i;
    char bytes[9];

    for (i = 0; i < 8; i++) {
        unsigned c = (x >> 28);
        x = x << 4;
        if (c >= 10) {
            c += 'A' - 10;
        } else {
            c += '0';
        }
        bytes[i] = c;
    }
    bytes[8] = '\0';
    print(bytes);
}

void printdec(unsigned x)
{
    if (x == 0) {
        print("0");
    } else {
        char bytes[16];
        unsigned i = 15;

        bytes[i] = '\0';
        while (x > 0) {
            unsigned div = x / 10;
            unsigned rem = x - (div * 10);
            i--;
            bytes[i] = '0' + (char) rem;
            x = div;
        }
        print(&bytes[i]);
    }
}

const char * debug_find_task(void * x)
{
    unsigned i;
    for (i = num_threads; i > 0; i--) {
        TCB * p = &thread_table[i];
        if (p->function == x) {
            return p->name;
        }
    }
    return 0;
}

void task_func_is_zero(void)
{
    print("interrupt: task func is zero!\n");
    exit(EXIT_NO_TASK);
}    

extern volatile unsigned cs_measure_sample;
extern unsigned measurement_patch_point[];
extern void cs_measure_ep(void);

void cs_measure(unsigned n, unsigned search_size,
            unsigned call_overhead, unsigned return_overhead) 
{
    unsigned i, j, a, b, c;
    unsigned cs_to_max = 0;
    unsigned cs_from_max = 0;
    unsigned cs_to, cs_from;
    unsigned save;

    num_threads = 0;
    for (i = 0; i < n; i++) {
        register_task("test", cs_measure_ep, 
                BIW(2, 2), 0, 1000000);
    }
    save = measurement_patch_point[0];
    measurement_patch_point[0] = 0x80000000;

    for (j = 0; j < search_size; j++) {
        set_trigger(-1);
        set_trigger(20);
        thread_table[0].function = crash;
        thread_table[0].running = 1;
        for (i = num_threads; i > 0; i--) {
            TCB * p = &thread_table[i];
            p->dynamic_release = -100;
            p->dynamic_deadline = -100;
            p->running = 0;
        }
        thread_table[1].dynamic_release = 10;
        thread_table[1].dynamic_deadline = 1000000;
        cs_measure_sample = 0;

        a = get_time();
        asm volatile("msrset  r5, 0x102\n");

        while (cs_measure_sample == 0) {}

        c = get_time();
        b = cs_measure_sample;
        set_trigger(1<<31);
        asm volatile("msrclr  r5, 0x102\n");

        cs_to = b - a;
        cs_from = c - b;

        if (j > 0) {
            if (cs_to_max < cs_to) {
                cs_to_max = cs_to;
            }
            if (cs_from_max < cs_from) {
                cs_from_max = cs_from;
            }
        }
    }

    measurement_patch_point[0] = save;

    print("# number of tasks ");
    printdec(n);
    print("\nCS1 = ");
    printdec((cs_to - call_overhead) * CLOCK_PERIOD);
    print("#ns\nCS2 = ");
    printdec((cs_from - return_overhead) * CLOCK_PERIOD);
    print("#ns\n");
    num_threads = 0;
}


